package java.util;

import java.util.function.Consumer;
import sun.misc.SharedSecrets;
import java.io.ObjectInputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.lang.reflect.Array;
import java.io.Serializable;

public class ArrayDeque<E> extends AbstractCollection<E> implements Deque<E>, Cloneable, Serializable
{
    transient Object[] elements;
    transient int head;
    transient int tail;
    private static final int MIN_INITIAL_CAPACITY = 8;
    private static final long serialVersionUID = 2340985798034038923L;
    static final /* synthetic */ boolean $assertionsDisabled;
    
    private static int calculateSize(int n) {
        int n2 = 8;
        if (n >= n2) {
            int n3 = n | n >>> 1;
            int n4 = n3 | n3 >>> 2;
            int n5 = n4 | n4 >>> 4;
            int n6 = n5 | n5 >>> 8;
            n2 = (n6 | n6 >>> 16);
            if (++n2 < 0) {
                n2 >>>= 1;
            }
        }
        return n2;
    }
    
    private void allocateElements(int n) {
        this.elements = new Object[calculateSize(n)];
    }
    
    private void doubleCapacity() {
        assert this.head == this.tail;
        int head = this.head;
        int length = this.elements.length;
        int n = length - head;
        int n2 = length << 1;
        if (n2 < 0) {
            throw new IllegalStateException("Sorry, deque too big");
        }
        Object[] elements = new Object[n2];
        System.arraycopy(this.elements, head, elements, 0, n);
        System.arraycopy(this.elements, 0, elements, n, head);
        this.elements = elements;
        this.head = 0;
        this.tail = length;
        return;
    }
    
    private <T> T[] copyElements(T[] array) {
        if (this.head < this.tail) {
            System.arraycopy(this.elements, this.head, array, 0, this.size());
        }
        else if (this.head > this.tail) {
            int n = this.elements.length - this.head;
            System.arraycopy(this.elements, this.head, array, 0, n);
            System.arraycopy(this.elements, 0, array, n, this.tail);
        }
        return array;
    }
    
    public ArrayDeque() {
        this.elements = new Object[16];
    }
    
    public ArrayDeque(int n) {
        this.allocateElements(n);
    }
    
    public ArrayDeque(Collection<? extends E> collection) {
        this.allocateElements(collection.size());
        this.addAll(collection);
    }
    
    @Override
    public void addFirst(E e) {
        if (e == null) {
            throw new NullPointerException();
        }
        this.elements[this.head = (this.head - 1 & this.elements.length - 1)] = e;
        if (this.head == this.tail) {
            this.doubleCapacity();
        }
        return;
    }
    
    @Override
    public void addLast(E e) {
        if (e == null) {
            throw new NullPointerException();
        }
        this.elements[this.tail] = e;
        if ((this.tail = (this.tail + 1 & this.elements.length - 1)) == this.head) {
            this.doubleCapacity();
        }
        return;
    }
    
    @Override
    public boolean offerFirst(E e) {
        this.addFirst(e);
        return true;
    }
    
    @Override
    public boolean offerLast(E e) {
        this.addLast(e);
        return true;
    }
    
    @Override
    public E removeFirst() {
        Object pollFirst = this.pollFirst();
        if (pollFirst == null) {
            throw new NoSuchElementException();
        }
        return pollFirst;
    }
    
    @Override
    public E removeLast() {
        Object pollLast = this.pollLast();
        if (pollLast == null) {
            throw new NoSuchElementException();
        }
        return pollLast;
    }
    
    @Override
    public E pollFirst() {
        int head = this.head;
        Object o = this.elements[head];
        if (o == null) {
            return null;
        }
        this.elements[head] = null;
        this.head = (head + 1 & this.elements.length - 1);
        return o;
    }
    
    @Override
    public E pollLast() {
        int tail = this.tail - 1 & this.elements.length - 1;
        Object o = this.elements[tail];
        if (o == null) {
            return null;
        }
        this.elements[tail] = null;
        this.tail = tail;
        return o;
    }
    
    @Override
    public E getFirst() {
        Object o = this.elements[this.head];
        if (o == null) {
            throw new NoSuchElementException();
        }
        return o;
    }
    
    @Override
    public E getLast() {
        Object o = this.elements[this.tail - 1 & this.elements.length - 1];
        if (o == null) {
            throw new NoSuchElementException();
        }
        return o;
    }
    
    @Override
    public E peekFirst() {
        return this.elements[this.head];
    }
    
    @Override
    public E peekLast() {
        return this.elements[this.tail - 1 & this.elements.length - 1];
    }
    
    @Override
    public boolean removeFirstOccurrence(Object o) {
        if (o == null) {
            return false;
        }
        int n = this.elements.length - 1;
        int head = this.head;
        Object obj;
        while ((obj = this.elements[head]) != null) {
            if (o.equals(obj)) {
                this.delete(head);
                return true;
            }
            head = (head + 1 & n);
        }
        return false;
    }
    
    @Override
    public boolean removeLastOccurrence(Object o) {
        if (o == null) {
            return false;
        }
        int n = this.elements.length - 1;
        int n2 = this.tail - 1 & n;
        Object obj;
        while ((obj = this.elements[n2]) != null) {
            if (o.equals(obj)) {
                this.delete(n2);
                return true;
            }
            n2 = (n2 - 1 & n);
        }
        return false;
    }
    
    @Override
    public boolean add(E e) {
        this.addLast(e);
        return true;
    }
    
    @Override
    public boolean offer(E e) {
        return this.offerLast(e);
    }
    
    @Override
    public E remove() {
        return this.removeFirst();
    }
    
    @Override
    public E poll() {
        return this.pollFirst();
    }
    
    @Override
    public E element() {
        return this.getFirst();
    }
    
    @Override
    public E peek() {
        return this.peekFirst();
    }
    
    @Override
    public void push(E e) {
        this.addFirst(e);
    }
    
    @Override
    public E pop() {
        return this.removeFirst();
    }
    
    private void checkInvariants() {
        assert this.elements[this.tail] == null;
        if (!ArrayDeque.$assertionsDisabled) {
            if ((this.head != this.tail) ? (this.elements[this.head] != null && this.elements[this.tail - 1 & this.elements.length - 1] != null) : (this.elements[this.head] == null)) {}
            throw new AssertionError();
        }
        assert this.elements[this.head - 1 & this.elements.length - 1] == null;
        return;
    }
    
    private boolean delete(int n) {
        this.checkInvariants();
        Object[] elements = this.elements;
        int n2 = elements.length - 1;
        int head = this.head;
        int tail = this.tail;
        int n3 = n - head & n2;
        int n4 = tail - n & n2;
        if (n3 >= (tail - head & n2)) {
            throw new ConcurrentModificationException();
        }
        if (n3 < n4) {
            if (head <= n) {
                System.arraycopy(elements, head, elements, head + 1, n3);
            }
            else {
                System.arraycopy(elements, 0, elements, 1, n);
                elements[0] = elements[n2];
                System.arraycopy(elements, head, elements, head + 1, n2 - head);
            }
            elements[head] = null;
            this.head = (head + 1 & n2);
            return false;
        }
        else {
            if (n < tail) {
                System.arraycopy(elements, n + 1, elements, n, n4);
                this.tail = tail - 1;
            }
            else {
                System.arraycopy(elements, n + 1, elements, n, n2 - n);
                elements[n2] = elements[0];
                System.arraycopy(elements, 1, elements, 0, tail);
                this.tail = (tail - 1 & n2);
            }
            return true;
        }
    }
    
    @Override
    public int size() {
        return this.tail - this.head & this.elements.length - 1;
    }
    
    @Override
    public boolean isEmpty() {
        return this.head == this.tail;
    }
    
    @Override
    public Iterator<E> iterator() {
        return new ArrayDeque.DeqIterator();
    }
    
    @Override
    public Iterator<E> descendingIterator() {
        return new ArrayDeque.DescendingIterator();
    }
    
    @Override
    public boolean contains(Object o) {
        if (o == null) {
            return false;
        }
        int n = this.elements.length - 1;
        int head = this.head;
        Object obj;
        while ((obj = this.elements[head]) != null) {
            if (o.equals(obj)) {
                return true;
            }
            head = (head + 1 & n);
        }
        return false;
    }
    
    @Override
    public boolean remove(Object o) {
        return this.removeFirstOccurrence(o);
    }
    
    @Override
    public void clear() {
        int i = this.head;
        int tail = this.tail;
        if (i != tail) {
            int n = 0;
            this.tail = n;
            this.head = n;
            int n2 = this.elements.length - 1;
            do {
                this.elements[i] = null;
                i = (i + 1 & n2);
            } while (i != tail);
        }
    }
    
    @Override
    public Object[] toArray() {
        return this.copyElements(new Object[this.size()]);
    }
    
    @Override
    public <T> T[] toArray(T[] array) {
        int size = this.size();
        if (array.length < size) {
            array = (Object[])(Object[])Array.newInstance(array.getClass().getComponentType(), size);
        }
        this.copyElements(array);
        if (array.length > size) {
            array[size] = null;
            return array;
        }
        return array;
    }
    
    public ArrayDeque<E> clone() {
        try {
            ArrayDeque<Object> arrayDeque = (ArrayDeque)super.clone();
            arrayDeque.elements = Arrays.copyOf(this.elements, this.elements.length);
            return arrayDeque;
        }
        catch (CloneNotSupportedException ex) {
            throw new AssertionError();
        }
    }
    
    private void writeObject(ObjectOutputStream objectOutputStream) throws IOException {
        objectOutputStream.defaultWriteObject();
        objectOutputStream.writeInt(this.size());
        for (int n = this.elements.length - 1, i = this.head; i != this.tail; i = (i + 1 & n)) {
            objectOutputStream.writeObject(this.elements[i]);
        }
    }
    
    private void readObject(ObjectInputStream objectInputStream) throws IOException, ClassNotFoundException {
        objectInputStream.defaultReadObject();
        int int1 = objectInputStream.readInt();
        SharedSecrets.getJavaOISAccess().checkArray(objectInputStream, Object[].class, calculateSize(int1));
        this.allocateElements(int1);
        this.head = 0;
        this.tail = int1;
        for (int i = 0; i < int1; ++i) {
            this.elements[i] = objectInputStream.readObject();
        }
    }
    
    @Override
    public Spliterator<E> spliterator() {
        return new ArrayDeque.DeqSpliterator(this, -1, -1);
    }
    
    static boolean access$200(ArrayDeque arrayDeque, int n) {
        return arrayDeque.delete(n);
    }
    
    private class DeqIterator implements Iterator<E>
    {
        private int cursor;
        private int fence;
        private int lastRet;
        
        private DeqIterator() {
            this.cursor = ArrayDeque.this.head;
            this.fence = ArrayDeque.this.tail;
            this.lastRet = -1;
        }
        
        @Override
        public boolean hasNext() {
            return this.cursor != this.fence;
        }
        
        @Override
        public E next() {
            if (this.cursor == this.fence) {
                throw new NoSuchElementException();
            }
            Object o = ArrayDeque.this.elements[this.cursor];
            if (ArrayDeque.this.tail != this.fence || o == null) {
                throw new ConcurrentModificationException();
            }
            this.lastRet = this.cursor;
            this.cursor = (this.cursor + 1 & ArrayDeque.this.elements.length - 1);
            return o;
        }
        
        @Override
        public void remove() {
            if (this.lastRet < 0) {
                throw new IllegalStateException();
            }
            if (ArrayDeque.access$200(ArrayDeque.this, this.lastRet)) {
                this.cursor = (this.cursor - 1 & ArrayDeque.this.elements.length - 1);
                this.fence = ArrayDeque.this.tail;
            }
            this.lastRet = -1;
            return;
        }
        
        @Override
        public void forEachRemaining(Consumer<? super E> consumer) {
            Objects.requireNonNull(consumer);
            Object[] elements = ArrayDeque.this.elements;
            int n = elements.length - 1;
            int fence = this.fence;
            int i = this.cursor;
            this.cursor = fence;
            while (i != fence) {
                Object o = elements[i];
                i = (i + 1 & n);
                if (o == null) {
                    throw new ConcurrentModificationException();
                }
                consumer.accept(o);
            }
        }
    }
    
    private class DescendingIterator implements Iterator<E>
    {
        private int cursor;
        private int fence;
        private int lastRet;
        
        private DescendingIterator() {
            this.cursor = ArrayDeque.this.tail;
            this.fence = ArrayDeque.this.head;
            this.lastRet = -1;
        }
        
        @Override
        public boolean hasNext() {
            return this.cursor != this.fence;
        }
        
        @Override
        public E next() {
            if (this.cursor == this.fence) {
                throw new NoSuchElementException();
            }
            this.cursor = (this.cursor - 1 & ArrayDeque.this.elements.length - 1);
            Object o = ArrayDeque.this.elements[this.cursor];
            if (ArrayDeque.this.head != this.fence || o == null) {
                throw new ConcurrentModificationException();
            }
            this.lastRet = this.cursor;
            return o;
        }
        
        @Override
        public void remove() {
            if (this.lastRet < 0) {
                throw new IllegalStateException();
            }
            if (!ArrayDeque.access$200(ArrayDeque.this, this.lastRet)) {
                this.cursor = (this.cursor + 1 & ArrayDeque.this.elements.length - 1);
                this.fence = ArrayDeque.this.head;
            }
            this.lastRet = -1;
            return;
        }
    }
    
    static final class DeqSpliterator<E> implements Spliterator<E>
    {
        private final ArrayDeque<E> deq;
        private int fence;
        private int index;
        
        DeqSpliterator(ArrayDeque<E> deq, int index, int fence) {
            this.deq = deq;
            this.index = index;
            this.fence = fence;
        }
        
        private int getFence() {
            int fence;
            if ((fence = this.fence) < 0) {
                int tail = this.deq.tail;
                this.fence = tail;
                fence = tail;
                this.index = this.deq.head;
            }
            return fence;
        }
        
        @Override
        public ArrayDeque.DeqSpliterator<E> trySplit() {
            int fence = this.getFence();
            int index = this.index;
            int length = this.deq.elements.length;
            if (index != fence && (index + 1 & length - 1) != fence) {
                if (index > fence) {
                    fence += length;
                }
                return new ArrayDeque.DeqSpliterator<E>(this.deq, index, this.index = (index + fence >>> 1 & length - 1));
            }
            else {
                return null;
            }
        }
        
        @Override
        public void forEachRemaining(Consumer<? super E> consumer) {
            if (consumer == null) {
                throw new NullPointerException();
            }
            Object[] elements = this.deq.elements;
            int n = elements.length - 1;
            int fence = this.getFence();
            int i = this.index;
            this.index = fence;
            while (i != fence) {
                Object o = elements[i];
                i = (i + 1 & n);
                if (o == null) {
                    throw new ConcurrentModificationException();
                }
                consumer.accept(o);
            }
            return;
        }
        
        @Override
        public boolean tryAdvance(Consumer<? super E> consumer) {
            if (consumer == null) {
                throw new NullPointerException();
            }
            Object[] elements = this.deq.elements;
            int n = elements.length - 1;
            this.getFence();
            int index = this.index;
            if (index != this.fence) {
                Object o = elements[index];
                this.index = (index + 1 & n);
                if (o == null) {
                    throw new ConcurrentModificationException();
                }
                consumer.accept(o);
                return true;
            }
            else {
                return false;
            }
        }
        
        @Override
        public long estimateSize() {
            int n = this.getFence() - this.index;
            if (n < 0) {
                n += this.deq.elements.length;
            }
            return (long)n;
        }
        
        @Override
        public int characteristics() {
            return 16720;
        }
    }
}
